@@ -77,6 +77,8 @@ class KeycloakOpenID:
     :param cert: An SSL certificate used by the requested host to authenticate the client.
         Either a path to an SSL certificate file, or two-tuple of
         (certificate file, key file).
+    :param max_retries: The total number of times to retry HTTP requests.
+    :type max_retries: int
     """
 
     def __init__(
@@ -90,6 +92,7 @@ class KeycloakOpenID:
         proxies=None,
         timeout=60,
         cert=None,
+        max_retries=1,
     ):
         """Init method.
 
@@ -114,6 +117,8 @@ class KeycloakOpenID:
             Either a path to an SSL certificate file, or two-tuple of
             (certificate file, key file).
         :type cert: Union[str,Tuple[str,str]]
+        :param max_retries: The total number of times to retry HTTP requests.
+        :type max_retries: int
         """
         self.client_id = client_id
         self.client_secret_key = client_secret_key
@@ -126,6 +131,7 @@ class KeycloakOpenID:
             verify=verify,
             proxies=proxies,
             cert=cert,
+            max_retries=max_retries,
         )
 
         self.authorization = Authorization()
@@ -251,7 +257,7 @@ class KeycloakOpenID:
         data_raw = self.connection.raw_get(URL_WELL_KNOWN.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
-    def auth_url(self, redirect_uri, scope="email", state=""):
+    def auth_url(self, redirect_uri, scope="email", state="", nonce=""):
         """Get authorization URL endpoint.
 
         :param redirect_uri: Redirect url to receive oauth code
@@ -260,6 +266,8 @@ class KeycloakOpenID:
         :type scope: str
         :param state: State will be returned to the redirect_uri
         :type state: str
+        :param nonce: Associates a Client session with an ID Token to mitigate replay attacks
+        :type nonce: str
         :returns: Authorization URL Full Build
         :rtype: str
         """
@@ -269,6 +277,7 @@ class KeycloakOpenID:
             "redirect-uri": redirect_uri,
             "scope": scope,
             "state": state,
+            "nonce": nonce,
         }
         return URL_AUTH.format(**params_path)
 
@@ -276,7 +285,7 @@ class KeycloakOpenID:
         self,
         username="",
         password="",
-        grant_type=["password"],
+        grant_type="password",
         code="",
         redirect_uri="",
         totp=None,
@@ -338,7 +347,7 @@ class KeycloakOpenID:
         )
         return raise_error_from_response(data_raw, KeycloakPostError)
 
-    def refresh_token(self, refresh_token, grant_type=["refresh_token"]):
+    def refresh_token(self, refresh_token, grant_type="refresh_token"):
         """Refresh the user token.
 
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
@@ -409,7 +418,7 @@ class KeycloakOpenID:
         """
         params_path = {"realm-name": self.realm_name}
         payload = {
-            "grant_type": ["urn:ietf:params:oauth:grant-type:token-exchange"],
+            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
             "client_id": self.client_id,
             "subject_token": token,
             "subject_token_type": subject_token_type,
@@ -816,7 +825,7 @@ class KeycloakOpenID:
         )
         return raise_error_from_response(data_raw, KeycloakPostError)
 
-    def device(self):
+    def device(self, scope: str = ""):
         """Get device authorization grant.
 
         The device endpoint is used to obtain a user code verification and user authentication.
@@ -831,11 +840,13 @@ class KeycloakOpenID:
         https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow
         https://github.com/keycloak/keycloak-community/blob/main/design/oauth2-device-authorization-grant.md#how-to-try-it
 
+        :param scope: Scope of authorization request, split with the blank space
+        :type scope: str
         :returns: Device Authorization Response
         :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
-        payload = {"client_id": self.client_id}
+        payload = {"client_id": self.client_id, "scope": scope}
 
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_DEVICE.format(**params_path), data=payload)
@@ -895,7 +906,7 @@ class KeycloakOpenID:
         data_raw = await self.connection.a_raw_get(URL_WELL_KNOWN.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
-    async def a_auth_url(self, redirect_uri, scope="email", state=""):
+    async def a_auth_url(self, redirect_uri, scope="email", state="", nonce=""):
         """Get authorization URL endpoint asynchronously.
 
         :param redirect_uri: Redirect url to receive oauth code
@@ -904,6 +915,8 @@ class KeycloakOpenID:
         :type scope: str
         :param state: State will be returned to the redirect_uri
         :type state: str
+        :param nonce: Associates a Client session with an ID Token to mitigate replay attacks
+        :type nonce: str
         :returns: Authorization URL Full Build
         :rtype: str
         """
@@ -913,6 +926,7 @@ class KeycloakOpenID:
             "redirect-uri": redirect_uri,
             "scope": scope,
             "state": state,
+            "nonce": nonce,
         }
         return URL_AUTH.format(**params_path)
 
@@ -920,7 +934,7 @@ class KeycloakOpenID:
         self,
         username="",
         password="",
-        grant_type=["password"],
+        grant_type="password",
         code="",
         redirect_uri="",
         totp=None,
@@ -982,7 +996,7 @@ class KeycloakOpenID:
         )
         return raise_error_from_response(data_raw, KeycloakPostError)
 
-    async def a_refresh_token(self, refresh_token, grant_type=["refresh_token"]):
+    async def a_refresh_token(self, refresh_token, grant_type="refresh_token"):
         """Refresh the user token asynchronously.
 
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
@@ -1053,7 +1067,7 @@ class KeycloakOpenID:
         """
         params_path = {"realm-name": self.realm_name}
         payload = {
-            "grant_type": ["urn:ietf:params:oauth:grant-type:token-exchange"],
+            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
             "client_id": self.client_id,
             "subject_token": token,
             "subject_token_type": subject_token_type,
@@ -1458,7 +1472,7 @@ class KeycloakOpenID:
         )
         return raise_error_from_response(data_raw, KeycloakPostError)
 
-    async def a_device(self):
+    async def a_device(self, scope: str = ""):
         """Get device authorization grant asynchronously.
 
         The device endpoint is used to obtain a user code verification and user authentication.
@@ -1473,11 +1487,13 @@ class KeycloakOpenID:
         https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow
         https://github.com/keycloak/keycloak-community/blob/main/design/oauth2-device-authorization-grant.md#how-to-try-it
 
+        :param scope: Scope of authorization request, split with the blank space
+        :type scope: str
         :returns: Device Authorization Response
         :rtype: dict
         """
         params_path = {"realm-name": self.realm_name}
-        payload = {"client_id": self.client_id}
+        payload = {"client_id": self.client_id, "scope": scope}
 
         payload = self._add_secret_key(payload)
         data_raw = await self.connection.a_raw_post(URL_DEVICE.format(**params_path), data=payload)
